/*
 * Copyright 1984, 1988, 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include "is_types.h"

#ifndef lint
static	char	sccsid[] = "@(#)is_date.c	1.2	11:48:49	1/29/92";
#endif

extern	int	db_errno;

static	int	doy[] = {	/* first day of month is this day of year */
		0, 31, 59, 90, 120, 151,
		181, 212, 243, 273, 304, 334 };

/*
 * Determine what the ordering is supposed to be for the date
 * values in a character date string.  There are three possible
 * ways to order the information.  We default to the standard
 * format used in the United States.
 */

#ifdef	ISAM_DMY
#undef	ISAM_MDY
#undef	ISAM_YMD
#endif

#ifdef	ISAM_MDY
#undef	ISAM_YMD
#endif

#if !defined(ISAM_MDY) && !defined(ISAM_DMY) && !defined(ISAM_YMD)
#define	ISAM_MDY
#endif

long	isam_date2int();
char	*isam_int2date();

/*
 * dt_isleap - is "year" a leap year
 *
 * dt_isleap() returns TRUE if the given year is a leap-year.
 * The rule is that any year which is evenly divisibly by 4
 * is a leap-year unless it is divisibly by 100 but not by 400.
 */

static int
dt_isleap (year)
int	year;
{
	if (year % 4 != 0)
		return (0);

	if (year % 100 != 0)
		return (1);

	if (year % 400 != 0)
		return (0);

	return (1);
}

static int
dt_days (month, year)
int	month;
int	year;
{
	static	int	months[12] =
				{ 31, 28, 31,
				  30, 31, 30,
				  31, 31, 30,
				  31, 30, 31 };

	if (month != 2)
		return (months[month - 1]);

	return (28 + dt_isleap (year));
}

static long
do_mdy (month, day, year)
int	month;
int	day;
int	year;
{
	long	julian;

	/*
	 * Normalize the year.  The year may be entered in the range
	 * of 0 .. 99 for the 20th Century, or as a full 4 digit year
	 * for any other century, up to 4000AD.
	 */

	if (year >= 1900)
		year -= 1900;
	else if (year >= 100)
		return (0);
	else if (year < 0)
		return 0;

	year += 1900;

	/*
	 * Validate the month and day parameters.  This makes certain
	 * the month and day are in bounds.
	 */

	if (year >= 4000 || month <= 0 || month > 12 ||
				day <= 0 || day > dt_days (month, year))
		return (db_errno = DB_INVALID), 0;

	/*
	 * Compute the Julian day, using 1/0/1900 as the epoch.  Start by
	 * computing the usual method, then correct for centuries not being
	 * leap-years unless they are divisible by 400.
	 */

	julian = (((year - 1900L) * 365L) + ((year - 1901L) / 4L));
	julian -= (year - 1901L) / 100L;
	julian += (year - 1601L) / 400L;

	/*
	 * Throw in the days in the preceding months and days in the current
	 * month to round things off.
	 */

	julian += doy[month - 1] + day;

	if (month > 2)
		julian += dt_isleap (year);

	return (julian);
}

/*
 * isam_date2int - convert date string to an integer
 *
 *	isam_date2int() converts a string of the form "MM/DD/[CC]YY" to a
 *	long integer in the range 1 to MAXLONG.  The epoch begins Jan 0, 1900.
 */

long
isam_date2int (cp, delimiter)
register char	*cp;
char delimiter;
{
	register int	i;
	register int	n1, n2, n3;

	for (n1 = 0;*cp >= '0' && *cp <= '9';cp++)
		n1 = n1 * 10 + *cp - '0';

	if (*cp++ != delimiter)
		return (db_errno = DB_INVALID), 0;

	for (n2 = 0;*cp >= '0' && *cp <= '9';cp++)
		n2 = n2 * 10 + *cp - '0';

	if (*cp++ != delimiter)
		return (db_errno = DB_INVALID), 0;

	for (n3 = 0;*cp >= '0' && *cp <= '9';cp++)
		n3 = n3 * 10 + *cp - '0';

	if (*cp != 0)
		return (db_errno = DB_INVALID), 0;

/*
 * The do_mdy() function requires the parameters in "month" "day" "year"
 * order, just like the name suggests ...
 *
 * So, we pick the correct calling sequence given the order the values
 * were placed in the string.
 */

#ifdef	ISAM_MDY
	return do_mdy (n1, n2, n3);
#endif
#ifdef	ISAM_DMY
	return do_mdy (n2, n1, n3);
#endif
#ifdef	ISAM_YMD
	return do_mdy (n2, n3, n1);
#endif
}

/*
 * isam_date2int - convert a long integer to a julian date string
 */

char *
isam_int2date (julian, delimiter)
long	julian;
int	delimiter;
{
	static	char	datebuf[11];
	int	year, month, day;
	int	i;
	char	*strcpy ();
	long	orig = julian;

	/*
	 * The earliest date is 1/1/1900.  Any date before that is
	 * considered to be an underflow.
	 */

	if (julian <= 0)
		return (db_errno = DB_INVALID),
			(strcpy (datebuf, "00/00/0000"));

	/*
	 * The latest date is 12/31/3999.  Any date after that is
	 * considered to be an overflow.
	 */

	if (julian > 767009L)
		return (db_errno = DB_INVALID),
			(strcpy (datebuf, "99/99/9999"));

	/*
	 * Figure out what year it is, and remove all the days up
	 * until the beginning of the year.
	 */

	year = (int) ((double) julian / 365.2424) + 1900;
	if ((julian -= do_mdy (1, 1, year) - 1) <= 0) {
		year--;
		julian = orig - do_mdy (1, 1, year) + 1;
	}
	if (julian > 366 || (! dt_isleap (year) && julian > 365)) {
		year++;
		julian = orig - do_mdy (1, 1, year) + 1;
	}

	/*
	 * Take the remaining days and turn it into a number of
	 * months by slicing off a month's worth.  Whatever is
	 * left becomes the number of days.
	 */

	for (month = 1;julian > dt_days (month, year);month++)
		julian -= dt_days (month, year);

	day = julian;

	/*
	 * Convert the final result into an output string.
	 */

	if (year < 2000)
		sprintf (datebuf, "%2.02d%c%2.02d%c%2.02d",
#ifdef	ISAM_MDY
			month, delimiter, day, delimiter, year - 1900);
#endif
#ifdef	ISAM_DMY
			day, delimiter, month, delimiter, year - 1900);
#endif
#ifdef	ISAM_YMD
			year - 1900, delimiter, month, delimiter, day);
#endif
	else
		sprintf (datebuf, "%2.02d%c%2.02d%c%4.04d",
#ifdef	ISAM_MDY
			month, delimiter, day, delimiter, year);
#endif
#ifdef	ISAM_DMY
			day, delimiter, month, delimiter, year);
#endif
#ifdef	ISAM_YMD
			year, delimiter, month, delimiter, day);
#endif

	return (datebuf);
}
