/*
 * Copyright 1991, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <math.h>
#include <ctype.h>
#ifdef BSD
#include <strings.h>
#define strchr index
#define strrchr rindex
#else
#include <string.h>
#include <memory.h>
#endif

#include "is_types.h"
#include "isam.h"

#ifndef lint
static	char	sccsid[] = "@(#)is_string.c	1.9	11:16:16	12/27/91";
#endif

extern	long	strtol ();
extern	double	strtod ();

/*
 * string2buf - convert a text string to a data buffer
 *
 * string2buf() converts a text string which was (presumably) created
 * with buf2string() back into a data buffer.  This routine understands
 * certain escape sequences and uses them to convert special characters
 * back to their normal format.
 */

char *
string2buf (file, string, delimiter, buffer)
file_t	*file;
char	*string;
int	delimiter;
char	*buffer;
{
	field_t	*fields;
	view_t	*view;
	char	*cp, *next;
	char	*ep;
	int	cnt;
	short	s;
	int	i;
	int	len;
	long	l;
	float	f;
	double	d, m;

	/*
	 * Point to the start of the data buffer
	 */

	cp = buffer;

	/*
	 * If the file has a view defined for it, the buffer will be
	 * considered to be the file view, otherwise the entire data
	 * record is used as the template for the buffer
	 */

	if (file->f_current_view) {
		cnt = 0;
		view = file->f_current_view;
#ifdef BSD
		bzero (buffer, view->v_size);
#else
		memset (buffer, 0, view->v_size);
#endif /* BSD */
	} else {
		cnt = 1;
		fields = &file->f_schema->s_fields[1];
#ifdef BSD
		bzero (buffer, file->f_record_size);
#else
		memset (buffer, 0, file->f_record_size);
#endif /* BSD */
		view = 0;
	}

	/*
	 * Loop through every field in the data buffer
	 */

	for (;(view && cnt < view->v_field_count) ||
			(! view && cnt < file->f_schema->s_field_count);cnt++) {

		/*
		 * Figure out which field is to be used for the next
		 * part of the string.
		 */

		if (view) {
			i = view->v_fields[cnt];
			cp = view->v_offsets[cnt] + buffer;
			fields = file->f_schema->s_fields + i;
		} else
			fields = file->f_schema->s_fields + cnt;

		/*
		 * Index fields are ignored completely
		 */

		if (fields->f_type == T_INDEX)
			continue;

		/*
		 * Check that there is still some data left.  It is
		 * permissible for a CHAR field to be blank, but no
		 * type data can be missing.
		 */

		if (! *string && fields->f_type != T_CHAR)
			return 0;

		/*
		 * Scan the string for a delimiter and mark where
		 * the next field starts.
		 */

		if (next = strchr (string, delimiter))
			*next++ = '\0';

		/*
		 * Take the current chunk of the string and convert
		 * it into the correct data type and copy it into the
		 * user's buffer.
		 */

		switch (fields->f_type) {
			case T_SHORT:

				/* 
				 * Convert the string to a short integer
				 */

				s = (short) strtol (string, &ep, 0);
				if (*ep)
					return 0;

				memcpy (cp, &s, sizeof s);
				cp += sizeof s;
				break;
			case T_INT:

				/*
				 * Convert the string to a long integer,
				 * then copy as the default "int" type.
				 */

				i = (int) strtol (string, &ep, 0);
				if (*ep)
					return 0;

				memcpy (cp, &i, sizeof i);
				cp += sizeof i;
				break;
			case T_DATE:

				/*
				 * Convert the string from "date" format
				 * to a long integer.
				 */

				l = (long) isam_date2int (string, '/');
				memcpy (cp, &l, sizeof l);
				cp += sizeof l;
				break;
			case T_LONG:
			case T_SERIAL:

				/*
				 * Convert the string to a long integer.
				 */

				l = (long) strtol (string, &ep, 0);
				if (*ep)
					return 0;

				memcpy (cp, &l, sizeof l);
				cp += sizeof l;
				break;
			case T_FLOAT:

				/*
				 * Convert the string to a single precision
				 * floating point number.
				 */

				f = (float) strtod (string, &ep);
				if (*ep)
					return 0;

				memcpy (cp, &f, sizeof f);
				cp += sizeof f;
				break;
			case T_DOUBLE:

				/*
				 * Convert the string to a double precision
				 * floating point number.
				 */

				d = strtod (string, &ep);
				if (*ep)
					return 0;

				memcpy (cp, &d, sizeof d);
				cp += sizeof d;
				break;
			case T_MONEY:

				/* 
				 * Convert the string to a double precision
				 * number, then scale by 100.  The result of
				 * that scaling will then be made into a
				 * whole number of cents.
				 */

				d = strtod (string, &ep);
				if (*ep)
					return 0;

				d *= 100.0;
				m = fmod (d, 1.0);
				if (d >= 0.0) {
					if (m >= 0.5)
						d = ceil (d);
					else
						d = floor (d);
				} else {
					if (-m >= 0.5)
						d = floor (d);
					else
						d = ceil (d);
				}
				memcpy (cp, &d, sizeof d);
				cp += sizeof d;
				break;
			case T_CHAR:

				/*
				 * Copy the string as is, converting
				 * any "\" escaped octal strings into
				 * their corresponding values.
				 */

#ifdef BSD
				bzero (cp, fields->f_size);
#else
				memset (cp, 0, fields->f_size);
#endif /* BSD */
				for (i = 0;i < fields->f_size;i++) {
					if (*string != '\\')
						cp[i] = *string++;
					else {
						cp[i] = (string[1] - '0') << 6;
						cp[i] += (string[2] - '0') << 3;
						cp[i] += string[3] - '0';
						string += 4;
					}
				}
				cp += fields->f_size;
				break;
		}

		/* 
		 * Skip to the next field in the string.  If this was
		 * the end, just point string to a '\0' character.
		 */

		if (! next)
			string = "\0";
		else
			string = next;
	}

	/*
	 * Return a pointer to the user's buffer
	 */

	return buffer;
}

/* 
 * buf2string - convert a file buffer to a text string
 *
 * buf2string() converts the binary data in a buffer into a printable
 * string.  Binary fields are converted into human-readable formats, and
 * text fields are copied as is, with special conversions being made for
 * certain control characters.
 */

char *
buf2string (file, string, delimiter, buffer)
file_t	*file;
char	*string;
char	delimiter;
char	*buffer;
{
	field_t	*fields;
	view_t	*view;
	char	*cp;
	char	*org = string;
	char	*ep;
	short	s;
	int	i, j;
	int	cnt;
	int	field_cnt = 0;
	long	l;
	float	f;
	double	d, m;

	/*
	 * Point to the beginning of the user's buffer
	 */

	cp = buffer;

	/*
	 * If the file has a view defined for it, the buffer only
	 * represents the fields defined for that view.  Otherwise
	 * all of the fields in the file are used.
	 */

	if (file->f_current_view) {
		cnt = 0;
		view = file->f_current_view;
	} else {
		cnt = 1;
		fields = &file->f_schema->s_fields[1];
		view = 0;
	}

	/*
	 * Loop through the fields in the buffer
	 */

	for (;(view && cnt < view->v_field_count) ||
			(! view && cnt < file->f_schema->s_field_count);cnt++) {

		/*
		 * Figure out which field is to be used to convert the
		 * next chunk of data in the buffer
		 */

		if (view) {
			i = view->v_fields[cnt];
			fields = file->f_schema->s_fields + i;
			cp = view->v_offsets[cnt] + buffer;
		} else
			fields = file->f_schema->s_fields + cnt;

		/*
		 * Skip over index fields
		 */

		if (fields->f_type == T_INDEX)
			continue;

		/*
		 * If this is not the first field, add a delimiter to the
		 * current location so this field can be separated from the
		 * rest of the data.
		 */

		if (field_cnt++)
			*string++ = delimiter;

		/*
		 * Convert the next chunk
		 */

		switch (fields->f_type) {
			case T_SHORT:
do_short:
				/*
				 * Treat the next chunk as a short integer
				 * and output the numerical value for it.
				 */

				memcpy (&s, cp, sizeof s);
				sprintf (string, "%d", s);
				cp += sizeof s;
				break;
			case T_INT:

				/*
				 * Decide whether int's are shorts or longs
				 * and treat them accordingly.
				 */

				if (sizeof (int) == sizeof (short))
					goto do_short;
				else
					goto do_long;
			case T_DATE:

				/*
				 * Treat the next chunk as a long integer
				 * and convert it into a date string.
				 */

				memcpy (&l, cp, sizeof l);
				strcpy (string, isam_int2date (l, '/'));
				cp += sizeof l;
				break;
			case T_LONG:
			case T_SERIAL:
do_long:
				/*
				 * Treat the next chunk as a long integer
				 * and output it as it's numerical value.
				 */

				memcpy (&l, cp, sizeof l);
				sprintf (string, "%ld", l);
				cp += sizeof l;
				break;
			case T_FLOAT:

				/*
				 * Treat the next chunk as a single precision
				 * floating point number and output it in
				 * exponential notation.
				 */

				memcpy (&f, cp, sizeof f);
				sprintf (string, "%.7e", f);
				cp += sizeof f;
				break;
			case T_DOUBLE:

				/*
				 * Treat the next chunk as a double precision
				 * floating point number and output it in
				 * exponential notation
				 */

				memcpy (&d, cp, sizeof d);
				sprintf (string, "%.17e", d);
				cp += sizeof d;
				break;
			case T_MONEY:

				/*
				 * Treat the next chunk as a double precision
				 * floating point number, scale by .01, and
				 * output with two fractional digits.
				 */

				memcpy (&d, cp, sizeof d);
				d /= 100.0;
				sprintf (string, "%.2f", d);
				cp += sizeof d;
				break;
			case T_CHAR:

				/*
				 * Treat this as a text string.  Certain
				 * characters must be output as octal values
				 * since they can confuse the conversion
				 * routine in the other direction.
				 */

				for (j = 0;cp[j] && j < fields->f_size;j++) {
					if (cp[j] != '\\' && cp[j] != delimiter
							&& isprint (cp[j]))
						*string++ = cp[j];
					else {
						sprintf (string,
							"\\%03o", cp[j]);
						string += 4;
					}
				}
				*string = '\0';
				cp += fields->f_size;
				break;
		}

		/*
		 * Skip over the newly added string.
		 */

		string += strlen (string);
	}

	/* 
	 * NUL terminate and return the new string to the user.
	 */

	*string++ = '\0';
	return org;
}
